require 'Util'

local gaugeforcemt = {}
gaugeforcemt._type = "GaugeForce"
gaugeforcemt.__index = gaugeforcemt

function GaugeForce(lattice, opts)
  local gf = {}
  gf.lattice = lattice
  gf.group = lattice.defaultGroup
  gf.nc = lattice.defaultNc
  tableCopyTo(gf, opts)
  gf.force = qopqdp.force(gf.nc, lattice.qdplat)
  gf.nupdate = 0
  clearStats(gf, "update")
  return setmetatable(gf,gaugeforcemt)
end

function gaugeforcemt.clone(self)
  local gf = tableCopy(self)
  gf.force = qopqdp.force()
  gf.force:set(self.force)
  return setmetatable(gf,gaugeforcemt)
end

function gaugeforcemt.set(self, opts)
  if type(opts)=="table" then
    local t = opts._type
    if t=="GaugeForce" then
      self.force:set(opts.force)
    elseif t==nil then
      tableCopyTo(self, opts)
    else
      error("unknown type for opts: %s\n", t)
    end
  else
    error("unknown type for opts: %s\n", type(opts))
  end
end

function gaugeforcemt.load(self, filename)
  self.force:load(filename)
  self.filename = filename
  self.metadata = nil -- FIXME: get from file
  self.nupdate = self.nupdate + 1
end

function gaugeforcemt.save(self, filename, opts)
  if not self.metadata.file then
    self.metadata.file =
      "<?xml version=\"1.0\"?>\n<note>generated by qhmc</note>\n"
  end
  local precision = opts.precision or 1
  self.force:save(filename, self.metadata.file, precision)
  self.filename = filename
end

function gaugeforcemt.update(self, force, eps)
  local t0 = clock()
  self.force:update(force.force, eps)
  updateStats(self, "update", {seconds=(clock()-t0)})
  self.nupdate = self.nupdate + 1
end
