-- simple vector class

Vector = {}
local vecmt = {}

local function VectorCall(f, t)
  return setmetatable(t,vecmt)
end
setmetatable(Vector,{__call=VectorCall})

function Vector.isvector(v)
  return getmetatable(v) == vecmt
end

function Vector.zero(d)
  local v = {}
  for i=1,d do v[i] = 0 end
  return Vector(v)
end

function Vector.standardBasis(d)
  local b = {}
  for i=1,d do
    local e = {}
    for j=1,d do e[j] = 0 end
    e[i] = 1
    b[i] = Vector(e)
  end
  return b
end

function Vector.lexIndex(lat, x)
  local k = x[#x]
  for i=#x-1,1,-1 do
    k = (k*lat[i]) + x[i]
  end
  return k
end

function Vector.lexCoord(lat, k)
  local x = {}
  for i=1,#lat do
    x[i] = k % lat[i]
    k = math.floor(k / lat[i])
  end
  return Vector(x)
end

function vecmt.__add(a,b)
  local r = {}
  for i=1,#a do r[i] = a[i] + b[i] end
  return Vector(r)
end

function vecmt.__sub(a,b)
  local r = {}
  for i=1,#a do r[i] = a[i] - b[i] end
  return Vector(r)
end

function vecmt.__mul(a,b)
  local r = {}
  local num,v = a,b
  if Vector.isvector(a) then num,v = b,a end
  for i=1,#v do r[i] = num*v[i] end
  return Vector(r)
end

function vecmt.__eq(a,b)
  if #a~=#b then return false end
  local r = true
  for i=1,#a do r = r and (a[i] == b[i]) end
  return r
end

vecmt.lexIndex = Vector.lexIndex
vecmt.lexCoord = Vector.lexCoord
