--package.path = (userpath or "") .. "/nobackup1/weinbe2/bin/qhmc_sigma3/?.lua;/nobackup1/weinbe2/bin/qhmc_sigma3/hmc/?.lua;" .. arg[0]:gsub("[^/]*.lua","?.lua;") .. package.path -- engaging
--package.path = (userpath or "") .. "/nfs/tmp2/weinberg/qhmc/?.lua;/nfs/tmp2/weinberg/qhmc/hmc/?.lua;" .. arg[0]:gsub("[^/]*.lua","?.lua;") .. package.path -- Vulcan
package.path = (userpath or "") .. "/projectnb/qcd/Staggered8f/Disconnected/qhmc/?.lua;/projectnb/qcd/Staggered8f/Disconnected/qhmc/hmc/?.lua;" .. arg[0]:gsub("[^/]*.lua","?.lua;") .. package.path --SCC
require 'common'
require 'gaugeact'
require 'asqtadact'
require 'hmc'

-- This is a modified version of one of James' scripts for doing hmc generations.
-- It has been changed to the point that I'm not quite sure which one it's based off,
-- but in any case, it is designed in a way to only perform measurements, despite the
-- fact it looks like it's set up to do HMC generation.

-- Parameters we need to set.
local config = :CONFIGNUM:;
local nx = 24
local nt = 48
local mass = 0.015--0.010
local inlat = "../configurations/f4plus8l24t48b40m015m100_" .. config .. "_scidac"
--local inlat = ":CONFIG:"
local gfix_prec = 1e-6
local gfix_max = 4000
local cg_prec = 1e-6
local cg_max = 4000
local src_start = 2
local src_num = 6

-- Stuff for dilution!
local num_src = 6 -- Use 4 sources.
local time_dil = nt -- set to 0 to not time dilute.
local color_dil = 3 -- set to 0 to not color dilute.
local space_dil = 2 -- set to 0 to not space even/odd dilute


-- General lattice parameters.
local beta = 4.0
local beta_a = -0.25 -- for adjoint gauge action
local u0 = 1.0
local nf = 4 -- We just need to set this to something. Doesn't matter!

-- These parameters matter!
local prec = 1 -- Use single precision for inversions.

-- Parameters not likely needed anymore, but I keep around.
local ntraj = 10
local tau = 1
local nsteps = 120
seed = seed or os.time()
local first = -1
local ffreq = 3

-- Specify what lattice to load.

local latsize = { nx, nx, nx, nt }
local vol = 1
printf("latsize =")
for k,v in ipairs(latsize) do vol=vol*v; printf(" %i",v); end
printf("\nvolume = %i\n", vol)
printf("beta = %g\n", beta)
printf("u0 = %g\n", u0)
printf("nf = %g\n", nf)
printf("mass = %g\n", mass)
printf("ntraj = %g\n", ntraj)
printf("tau = %g\n", tau)
printf("nsteps = %g\n", nsteps)
printf("seed = %i\n", seed)

p = {}
p.latsize = latsize
p.vol = vol
p.beta = beta
p.seed = seed
p.gaugeact = {type="plaquette_adjoint", adjFac=beta_a}
--p.gaugeact = {type="plaquette", u0=u0}
--p.gaugeact = {type="symanzik_1loop", u0=u0}

-- The following chunk of functions for HMC still exist because
-- I never got around to deleting them and making sure it didn't break
-- anything. None of them should be relevant for measurements. 

function setpseudo(rhmc, mf, mb)
  local s1 = 4*mf*mf
  if mb then -- term (A+4mb^2)/(A+4mf^2)
    local s2 = 4*mb*mb
    local sr = math.sqrt(s1*s2)
    local c1 = s1 + s2 - 2*sr
    rhmc[#rhmc+1] = {
      --[[GR = {
	{ {1}, {sr-s2, s2} },
	{ {math.sqrt(c1), s2}, allfaceven=0, allfacodd=1 }
      },--]]
      GR = {
	{ {2, s2}, allfaceven=2*mf, allfacodd=1, allmass2=mb }
      },
      FA = {
	{ {1} },
	{ {math.sqrt(s2-s1), s1}, allfaceven=2*mf, allfacodd=1 }
      },
      MD = { {1}, {s2-s1, s1} }
    }
  else -- term 1/(A+4mf^2)
    rhmc[#rhmc+1] = {
      GR = {{ {1}, allfaceven=2*mf, allfacodd=1 }},
      FA = {{ {1,s1}, allfaceven=2*mf, allfacodd=1 }},
      MD = { {1,s1} }
    }
  end
end

local rhmc = {}
setpseudo(rhmc, mass)
--setpseudo(rhmc, mass2, mass)

local smear = {}

function projRat(s)
  local o = s.coeffs.one_link
  s.coeffs.one_link = nil
  return { type="projectRat", rho=1/o, smear=s }
  --return { type="projectRat", rho=4, smear=s }
end

-- I'm just going to trust these.
scoeffs = asqtad_coeffs(u0)
scoeffs.three_staple = -scoeffs.three_staple
scoeffs.seven_staple = -scoeffs.seven_staple
scoeffs.one_link = scoeffs.one_link + 3*scoeffs.naik
scoeffs.naik = nil
scoeffs.one_link = scoeffs.one_link + 6*scoeffs.lepage
scoeffs.lepage = nil

-- This one is still important! It specifies the nHYP smearing convention.
-- The order of coefficients is non-standard (it's backwards from most
-- conventions I've seen), but I'm verified that this order is necessary
-- to reproduce what MILC gives us.
smear[#smear+1] = { type="hyp", alpha={0.4,0.5,0.5} }
myprint("smear = ", smear, "\n")

-- This is to specify that we're reeeeally doing Staggered with nHYP smearing,
-- not a full blown ASQTAD improvement. You'd specify the coefficients of three
-- link, five link, etc terms here for ASQTAD. 
coeffs = { one_link=1 }

local act = {}
act.g = gaugeact(p)
act.f = asqtadact(act.g, {smear=smear,coeffs=coeffs,rhmc=rhmc})

-- Load the lattice we want to perform measurements on.
G = act.g:gaugeNew(act)
last = ntraj * tau
G:load(inlat)
--G:unit(); -- For debugging.


-- Reunitarize. 
do
  local devavg,devmax = G.g:checkSU()
  printf("unitarity deviation avg: %g  max: %g\n", devavg, devmax)
  G.g:makeSU()
  devavg,devmax = G.g:checkSU()
  printf("new unitarity deviation avg: %g  max: %g\n", devavg, devmax)
end


function measure(G)
  local t0 = clock()
  local ps,pt = G:plaq()
  printf("plaq ss: %-8g  st: %-8g  tot: %g\n", ps, pt, 0.5*(ps+pt))

  local nd = #G.a.latsize
  for i=1,nd do
    for j=i+1,nd do
      local lr,li = G.g:loop({i,j,-i,-j})
      printf("plaq%i%i:  %g\t%g\n", i, j, lr, li)
    end
  end

  local plpath = rep(-nd, G.a.latsize[nd])
  plpr,plpi = G.g:loop(plpath)
  printf("ploop:  %g\t%g\n", plpr, plpi)

  t0 = clock() - t0
  printf("meas time: %g\n", t0)

  io.stdout:flush()
end


-- Make some sources! 0 means U(1) noise.

eta_full = act.f:disc_get_sources(num_src, 0)
local t0 = clock();
phi_full = act.f:disc_get_propagators(eta_full, time_dil, color_dil, space_dil, G, mass, cg_prec, {prec = prec, restart = cg_max})
t0 = clock() - t0;
printf("Inversion time: %f\n", t0);
io.stdout:flush()

-- Get pbps!
t0 = clock();
local pbps = act.f:disc_pbp(eta_full, phi_full, time_dil, color_dil, space_dil, mass)
t0 = clock() - t0;
printf("Pbp contract time: %f\n", t0);

printf("BEGIN_PBP\n");
for i=1,#(pbps.unimp) do
   --printf("%i %.15e %.15e %.15e\n", i, pbps_other[i], pbps.unimp[i], pbps.isimp[i])
   printf("%i %.15e %.15e\n", i, pbps.unimp[i], pbps.isimp[i])
end
printf("END_PBP\n");
io.stdout:flush()

t0 = clock();
local pbps2 = act.f:disc_pbp_timeslice(eta_full, phi_full, time_dil, color_dil, space_dil, mass)
t0 = clock() - t0;
printf("Pbppart contract time: %f\n", t0);

printf("BEGIN_PBPPART\n");
for i=1,#(pbps2.unimp) do
   for t=1,#(pbps2.unimp[i]) do
     printf("%i\t%i\t%.15e\t%.15e\n", i, t, pbps2.unimp[i][t], pbps2.isimp[i][t])
   end
end

printf("END_PBPPART\n");
io.stdout:flush()

t0 = clock();
things = eta_full[1]:stoch_conn(eta_full, phi_full, time_dil, color_dil, space_dil);
t0 = clock() - t0;
printf("Correlator contract time: %f\n", t0);

printf("T\tScalar\n");
printf("BEGIN_SPECTRUM\n");

for i = 1,nt do
	printf("%i\t%.15e\n", i-1, things[i].r, things[i].i);
end

printf("END_SPECTRUM\n");
io.stdout:flush()

things = {};

-- Look at condensings of pbppart

if (time_dil ~= 0 and color_dil ~= 0 and space_dil ~= 0) then
	-- First, see what happens when we get rid of space e/o.
	local phi_reduced = {};
	for i=1,num_src do
		for j=1,(time_dil*color_dil) do
			phi_reduced[(i-1)*(time_dil*color_dil)+j] = act.f.h:quark();
			phi_reduced[(i-1)*(time_dil*color_dil)+j]:zero();
			phi_reduced[(i-1)*(time_dil*color_dil)+j]:combine({phi_full[(i-1)*nt*6+(j-1)*2+1],phi_full[(i-1)*nt*6+(j-1)*2+2]},{1,1});
		end
	end
	
	t0 = clock();
	local pbps_tc = act.f:disc_pbp_timeslice(eta_full, phi_reduced, time_dil, color_dil, 0, mass)
	t0 = clock() - t0;
	printf("Pbppart tc contract time: %f\n", t0);

	printf("BEGIN_PBPPART_TC\n");
	for i=1,#(pbps_tc.unimp) do
	   for t=1,#(pbps_tc.unimp[i]) do
		 printf("%i\t%i\t%.15e\t%.15e\n", i, t, pbps_tc.unimp[i][t], pbps_tc.isimp[i][t])
	   end
	end
	printf("END_PBPPART_TC\n");
	
	t0 = clock();
	things = eta_full[1]:stoch_conn(eta_full, phi_reduced, time_dil, color_dil, 0);
	t0 = clock() - t0;
	printf("Correlator tc contract time: %f\n", t0);

	printf("T\tScalar\n");
	printf("BEGIN_SPECTRUM_TC\n");

	for i = 1,nt do
		printf("%i\t%.15e\n", i-1, things[i].r, things[i].i);
	end

	printf("END_SPECTRUM_TC\n");
	
	--table.setn(phi_reduced, num_src*time_dil*space_dil);
	
	collectgarbage();
	
	local phi_reduced2 = {};
	for i=1,num_src do
		for j=1,(time_dil) do
			for k=1,space_dil do
				phi_reduced2[(i-1)*(time_dil*space_dil)+(j-1)*space_dil+k] = act.f.h:quark();
				phi_reduced2[(i-1)*(time_dil*space_dil)+(j-1)*space_dil+k]:zero();
				phi_reduced2[(i-1)*(time_dil*space_dil)+(j-1)*space_dil+k]:combine({phi_full[(i-1)*nt*6+(j-1)*6+0+k],phi_full[(i-1)*nt*6+(j-1)*6+2+k],phi_full[(i-1)*nt*6+(j-1)*6+4+k]},{1,1,1});
			end
		end
	end
	
	t0 = clock();
	local pbps_ts = act.f:disc_pbp_timeslice(eta_full, phi_reduced2, time_dil, 0, space_dil, mass)
	t0 = clock() - t0;
	printf("Pbppart ts contract time: %f\n", t0);

	printf("BEGIN_PBPPART_TS\n");
	for i=1,#(pbps_ts.unimp) do
	   for t=1,#(pbps_ts.unimp[i]) do
		 printf("%i\t%i\t%.15e\t%.15e\n", i, t, pbps_ts.unimp[i][t], pbps_ts.isimp[i][t]);
	   end
	end
	printf("END_PBPPART_TS\n");
	
	t0 = clock();
	things = eta_full[1]:stoch_conn(eta_full, phi_reduced2, time_dil, 0, space_dil);
	t0 = clock() - t0;
	printf("Correlator ts contract time: %f\n", t0);

	printf("T\tScalar\n");
	printf("BEGIN_SPECTRUM_TS\n");

	for i = 1,nt do
		printf("%i\t%.15e\n", i-1, things[i].r, things[i].i);
	end

	printf("END_SPECTRUM_TS\n");
	
	collectgarbage();
	
	--table.setn(phi_reduced, num_src*time_dil);
	
	local phi_reduced3 = {};
	for i=1,num_src do
	    for j=1,time_dil do
			phi_reduced3[(i-1)*nt+j] = act.f.h:quark();
			phi_reduced3[(i-1)*nt+j]:zero();
		    phi_reduced3[(i-1)*nt+j]:combine({phi_full[(i-1)*nt*6+(j-1)*6+1],phi_full[(i-1)*nt*6+(j-1)*6+2],phi_full[(i-1)*nt*6+(j-1)*6+3],phi_full[(i-1)*nt*6+(j-1)*6+4],phi_full[(i-1)*nt*6+(j-1)*6+5],phi_full[(i-1)*nt*6+(j-1)*6+6]},{1,1,1,1,1,1});
		end
	end
	
	t0 = clock();
	local pbps_t = act.f:disc_pbp_timeslice(eta_full, phi_reduced3, time_dil, 0, 0, mass)
	t0 = clock() - t0;
	printf("Pbppart t contract time: %f\n", t0);

	printf("BEGIN_PBPPART_T\n");
	for i=1,#(pbps_t.unimp) do
	   for t=1,#(pbps_t.unimp[i]) do
		 printf("%i\t%i\t%.15e\t%.15e\n", i, t, pbps_t.unimp[i][t], pbps_t.isimp[i][t]);
	   end
	end
	printf("END_PBPPART_T\n");

	t0 = clock();
	things = eta_full[1]:stoch_conn(eta_full, phi_reduced3, time_dil, 0, 0);
	t0 = clock() - t0;
	printf("Correlator t contract time: %f\n", t0);
	
	printf("BEGIN_SPECTRUM_T\n");

	for i = 1,nt do
		printf("%i\t%.15e\n", i-1, things[i].r, things[i].i);
	end

	printf("END_SPECTRUM_T\n");
	
else
	t0 = clock();
	things = eta_full[1]:stoch_conn(eta_full, phi_full, time_dil, color_dil, space_dil)
	t0 = clock() - t0;
	printf("Correlator contract time: %f\n", t0);
end
			


--t0 = clock();
--things = act.f:disc_get_many_corr(eta_full, phi_full, time_dil, color_dil, space_dil)
--t0 = clock() - t0;
--printf("Correlator contract time: %f\n", t0);


io.stdout:flush()

