function cgmss2(xt0, op, opts, subset)
  local maxits = opts.restart or 500
  local xt1 = xt0[1]
  local xt = {}
  local ps,shiftindex = {},{}
  for i=1,#xt0 do
    local xti = xt0[i]
    xt[i] = xti
    local sh = xti.shift
    local idx = shiftindex[sh]
    if idx==nil then
      idx = #ps+1
      ps[idx] = {
	shift = sh,
	nsrc = 0,
	p = xt0[1].src:clone(),
	Ap = xt0[1].x:clone(),
	beta = 0
      }
    end
    local psi = ps[idx]
    xti.psi = psi
    psi.nsrc = psi.nsrc + 1
  end
  local ps1 = xt1.psi
  local nx,np = #xt,#ps
  local r,p,Ap,shift = xt1.r,ps1.p,ps1.Ap,xt1.shift
  local Ar,Apo = xt1.src:clone(subset),xt1.src:clone(subset)
  local rsq,oldrsq = xt1.rsq
  local rsqstop = xt1.rsqstop
  local its = 0
  while true do
    Apo,Ap = Ap,Apo
    ps1.Ap = Ap
    op(Ap, p)
    Ap:combine({Ap,p}, {1,shift}, subset)
    its = its + 1
    ps1.pAp = p:Re_dot(Ap, subset)
    if ps1.pAp<=0 then break end  -- loss of precision in calculating pAp

    Ar:combine({Ap,Apo},{1,-ps1.beta},subset)
    for i=2,np do
      local psi = ps[i]
      local cr = psi.shift - shift
      psi.Ap:combine({psi.Ap,Ar,r},{psi.beta,1,cr},subset)
      --op(Ap[i], p[i])
      --Ap[i]:combine({Ap[i],p[i]}, {1,shifts[i]}, subset)
      psi.pAp = psi.p:Re_dot(psi.Ap, subset)
      if psi.pAp<=0 then goto exit end -- loss of precision in calculating pAp
    end

    xt1.alpha = rsq / ps1.pAp
    for i=2,nx do
      local xti = xt[i]
      xti.alpha = xti.psi.p:dot(xti.r,subset)/xti.psi.pAp
    end

    for i=1,nx do
      local xti = xt[i]
      xti.x:combine({xti.x,xti.psi.p},{1,xti.alpha}, subset)
      xti.r:combine({xti.r,xti.psi.Ap},{1,-xti.alpha}, subset)
    end
    oldrsq = rsq
    rsq = r:norm2(subset)
    --printf("%i\t%g\n", its, rsq)
    if its >= maxits or rsq < rsqstop then break end

    local xtn = xt[nx]
    xtn.rsq = xtn.r:norm2(subset)
    if xtn.rsq<xtn.rsqstop then
      local psi = xtn.psi
      psi.nsrc = psi.nsrc - 1
      if ps[np].nsrc == 0 then np = np - 1 end
      xtn.iterations = its
      nx = nx-1
    end

    ps1.beta = rsq / oldrsq
    for i=2,np do
      local psi = ps[i]
      psi.beta = -psi.Ap:Re_dot(r)/psi.pAp
    end
    for i=1,np do
      local psi = ps[i]
      psi.p:combine({psi.p,r},{psi.beta,1},subset)
    end
  end
  ::exit::
  for i=1,nx do xt[i].iterations = its end
  return {iterations=its, flops=0}
end

-- t = {[1]={x=x,src=src,shift=shift,resid=resid},...}
function cgmss(xt0, op, opts, subset)
  local flops,its = 0,0
  local srcsq,xt,imin = {},{},0
  for i=1,#xt0 do
    local xt0i = xt0[i]
    local src,x,shift,resid = xt0i.src,xt0i.x,xt0i.shift,xt0i.resid
    local ssq = srcsq[src]
    if ssq==nil then
      ssq = src:norm2(subset)
      srcsq[src] = ssq
    end
    local r,rsq,xx,rr
    if xt0i.useGuess then -- FIXME: should minimize
      r = src:clone(subset)
      op(r, x)
      r:combine({r,x,src},{-1,-shift,1},subset)
      rsq = r:norm2(subset)
      xx = x:clone(subset)
      xx:zero(subset)
      xt0i.iterations = 1
      its = its + 1
      --flops = flops + ???
    else
      x:zero(subset)
      r = src
      rsq = ssq
      xx = x
      xt0i.iterations = 0
    end
    rr = r:clone(subset)
    xt0i.ssq = ssq
    xt0i.rsq = rsq
    local rsqstop = resid*resid*ssq
    xt0i.rsqstop = rsqstop
    if rsq>rsqstop then
      xt[#xt+1] = {x=xx,src=r,shift=shift,rsqstop=rsqstop,r=rr,rsq=rsq,i=i}
      if imin==0 or shift<xt[imin].shift then imin = #xt end
    end
  end

  while #xt>0 do
    table.sort(xt,function(a,b) return a.shift<b.shift end)
    --if imin>1 then xt[1],xt[imin] = xt[imin],xt[1] end
    local info = cgmss2(xt, op, opts, subset)
    its = its + info.iterations
    flops = flops + info.flops

    local xtn = {}
    imin = 0
    for i=1,#xt do
      local xti = xt[i]
      local xt0i = xt0[xti.i]
      local x,x0 = xti.x,xt0i.x
      -- update solution
      if x~=x0 then
	x0:combine({x0,x},{1,1},subset)
      end
      local r,src0 = xti.r,xt0i.src
      op(r, x0)
      r:combine({r,x0,src0},{-1,-xt0i.shift,1})
      xt0i.rsq = r:norm2(subset)
      xt0i.iterations = xt0i.iterations + xti.iterations
      printf("%i\t%i\t%g\n", i, xt0i.iterations, xt0i.rsq/xt0i.rsqstop)
      if xt0i.rsq>xt0i.rsqstop then
	xtn[#xtn+1] = xti
	if x==x0 then xti.x = x:clone() end
	xti.x:zero()
	xti.src,xti.r = r,xti.src
	if xti.r==src0 then xti.r = r:clone() end
	xti.rsq = xt0i.rsq
	if imin==0 or xtn[#xtn].shift < xtn[imin].shift then imin = #xtn end
      end
    end
    xt = xtn
  end
  return {iterations=its,flops=flops}
end

function cgms(x, src, op, shifts, resid, opts, subset)
  local xt = {}
  for i=1,#x do
    xt[i] = {x=x[i],src=src,shift=shifts[i],resid=resid}
  end
  do
    local t = {x=x[1]:clone(),src=src:clone(),shift=shifts[1],resid=resid}
    --t.src:random()
    --t.src:combine({t.src,src},{0.001,1},subset)
    t.src:point({0,0,0,0},0,0,1,0)
    xt[#xt+1] = t
    t.x = t.x:clone()
    t.src = t.src:clone()
    t.src:point({0,0,0,0},0,1,1,0)
    xt[#xt+1] = t
  end
  local info = cgmss(xt, op, opts, subset)
  return {its=info.iterations,flops=info.flops}
end
