require 'common'
require 'gaugeact'

--trace(true)
latsize = { 4, 4, 4, 8 }
mass = 0.0008
prec = 2
restart = 10000
maxIter = 10000
resid = 1e-12
opts = { prec=prec, restart=restart, max_restarts=5, max_iter=maxIter, mixed_rsq=0 }
gaugePrecision = 2
betaG = 6
betaE = 1

--qopqdp.defaultNc(4); fn = nil
Nc = qopqdp.defaultNc()
printf("Nc = %d\n", Nc)

L = qopqdp.lattice(latsize)
qopqdp.profile(profile or 0)
qopqdp.verbosity(0)

seed = 987654321
seed = seed or os.time()
qopqdp.seed(seed)

function getplaq(g)
  local ps,pt = g:action{plaq=1}
  local lat = qopqdp.lattice()
  local nd,vol = #lat,1
  for i=1,nd do vol=vol*lat[i] end
  local s = 0.25*nd*(nd-1)*vol
  printf("plaq ss: %-8g  st: %-8g  tot: %-8g\n", ps/s, pt/s, 0.5*(ps+pt)/s)
end

g0 = qopqdp.gauge(gaugePrecision, Nc)
if fn then g0:load(fn)
else
  --g0:unit()
  g0:random()
end
getplaq(g0)
g0:heatbath(10, 1, 1, betaG, {plaq=1})
getplaq(g0)

-- coulomb(j_decay, error, max iterations, overrelaxation param)
-- note that here 0->x, ..., 3->t
--g:coulomb(3, 1e-7, 1000, 1.2)
--getplaq(g)

e0 = qopqdp.gauge(gaugePrecision, 1)
--e0:unit()
e0:random()
getplaq(e0)
e0:heatbath(10, 1, 1, betaE, {plaq=1})
getplaq(e0)

g = qopqdp.gauge(gaugePrecision, Nc)
qopqdp.smear({g}, {e0,g0}, {type="product",adj={false,false}})

function test(h)
  local pt = {0,0,0,0}
  printf("src point: %i %i %i %i\n", pt[1], pt[2], pt[3], pt[4])
  local src = h:quark()
  local dest = h:quark()
  local check = h:quark()
  src:point(pt,1,1)
  dest:zero()

  local t0 = qopqdp.dtime()
  h:solve({dest}, src, {mass}, resid, "all", opts)
  local dt = qopqdp.dtime() - t0
  local mf = 1e-6 * h:flops() / dt
  h:D(check, dest,mass)
  check:combine({check,src},{1,-1})
  local rsq = check:norm2()
  printf("its: %g  secs: %g  Mflops: %g  rsq: %g\n", h:its(), dt, mf, rsq)
  local p = dest:norm2("timeslices")
  for i=1,#p do
    printf("%i\t%g\n", i-1, p[i])
  end
end

--[[
h = qopqdp.hisq()
--h = qopqdp.asqtad()
--h:printcoeffs()

h:set(g0, 1, 0, prec)
--h:set(g0, prec)
test(h)

h:set(g, 1, 0, prec)
--h:set(g, prec)
test(h)
--]]

--he = qopqdp.asqtad(1)
--he:set(e0, prec)
he = qopqdp.hisq(1)
he:set(e0, 1, 0, prec)
se = he:quark()
de = he:quark()
--se:point({0,0,0,0,},1,1)
se:random()
print(se:norm2())
he:D(de,se,mass)
print(de:norm2())
test(he)
