require 'Lattice'
require 'Action'
require 'Evolver'

local nx = 4
local nt = 8
local beta3 = 6
local beta1 = 10

L = Lattice{nx,nx,nx,nt}
L:Seed(987654321)

G3 = L:GaugeField{group="SU",nc=3}
G3:Set("unit")
--G3:Load("lattice")

G1 = L:GaugeField{group="U",nc=1}
G1:Set("unit")
--G1:Load("lattice")

GA3 = Action{kind="gauge",style="plaquette",beta=beta3,field=G3}
GA1 = Action{kind="gauge",style="plaquette",beta=beta1,field=G1}

local Gmult = {}
Gmult.type = {type="product",adj={false,false}}
function Gmult.transform(sg, g)
  qopqdp.smear({sg[1].field}, {g[1]:GetField(),g[2]:GetField()}, Gmult.type)
end
function Gmult.chainrule(f, cf, sg, g)
  qopqdp.smearChain({f[1].field,f[2].field}, {cf[1].field}, {sg[1].field},
		    {g[1].field,g[2].field}, Gmult.type)
end

G = L:GaugeField{group="U",nc=3}
SetTransform({G}, {G3,G1}, Gmult)

M3 = G3:Momentum()
M1 = G1:Momentum()

MA3 = Action{kind="momentum",momentum=M3}
MA1 = Action{kind="momentum",momentum=M1}

M3:Random()
printf("M3: %s: %g\n", M3.group, M3:Norm2())
M1:Random()
printf("M1: %s: %g\n", M1.group, M1:Norm2())

I3G = Evolver{kind="md",style="leapfrog",action=GA3,
	      field=G3,momentum=M3,tau=1,nSteps=40}

local ss,st = G3:GetField():action({plaq=1})
printf("plaq: %g\t%g\n", ss, st)
printf("GA3: %g\t%g\n", GA3:Action(), GA3.act0)
printf("G3n: %g\n", G3:Norm2())
I3G:Reset()
I3G:Run()
local ss,st = G3:GetField():action({plaq=1})
printf("plaq: %g\t%g\n", ss, st)
printf("GA3: %g\t%g\n", GA3:Action(), GA3.act0)
printf("GA3: %g\n", GA3:Action())
printf("G3n: %g\n", G3:Norm2())
I3G:Reset()
I3G:Run()
local ss,st = G3:GetField():action({plaq=1})
printf("plaq: %g\t%g\n", ss, st)
printf("GA3: %g\t%g\n", GA3:Action(), GA3.act0)
printf("GA3: %g\n", GA3:Action())



os.exit(1)
M = G:Momentum()
MA = Action{kind="momentum",momentum=M}

--MHB = Evolver{kind="hb",action=MA,field=M}
I = Evolver{kind="md",style="leapfrog",action=GA3,
            field=G,momentum=M,tau=1,nSteps=40}

--MC = Evolver{kind="sequence",evolvers={MHB,I}}
E = Evolver{kind="mc",markov=I,actions={MA,GA3},fields={G}}
--E = Evolver{kind="metropolis",markov=I,actions={MA,GA},fields={G}}

printf("action: %g\n", GA:Action())
E:Run()
printf("action: %g\n", GA:Action())

myprint(E.oldActions,"\n")
myprint(E.newActions,"\n")
myprint(E.mcRand,"\n")
