require 'rat'
-- r(x) ~ (x+m)^p x=[a,b]
-- r(sy-m) ~ (sy)^p y=[(a+m)/s,(b+m)/s], x = sy-m, s = b+m
-- q(y) = r(sy-m)/s^p ~ y^p y=[(a+m)/(b+m),1]
-- r(x) = s^p q((x+m)/s)

function ratcoeffs(f, d, a, b)
  local ni = 40
  local xi = logspace(a, b, ni)
  local fi = {}
  local wfi = {}
  local m = matrix(ni, 2*d+1)
  local wm = matrix(ni, 2*d+1)
  local s
  local pc,qc,pi,qi,yi = {},{},{},{},{}
  for i=1,ni do
    --local x = xi[i]
    local x = bc.number(xi[i])
    local fx = f(x)
    fi[i] = fx
    local xp = bc.number(1)
    m:set(i,1, xp)
    for j=1,d do
      xp = xp * x
      m:set(i,j+1, xp)
      m:set(i,j+d+1, -xp*fx)
    end
  end
  local function updatec(wdef)
    for i=1,ni do
      local w = wdef(i)
      wfi[i] = w*fi[i]
      for j=1,2*d+1 do
	wm:set(i,j, w*m(i,j))
      end
    end
    s = wm:solve(wfi)
    for i=1,d+1 do pc[i] = s[i] end
    qc[1] = bc.number(1)
    for i=2,d+1 do qc[i] = s[d+i] end
    --print("pc ="); for i=1,#pc do print(" ",i,pc[i]) end
    --print("qc ="); for i=1,#qc do print(" ",i,qc[i]) end
    local sumsq,max = 0,0
    for i=1,ni do
      pi[i] = polyval(pc,xi[i])
      qi[i] = polyval(qc,xi[i])
      yi[i] = pi[i]/qi[i]
      local df = fi[i] - yi[i]
      sumsq = sumsq + df*df
      local adf = (df:isneg() and -df) or df
      if adf>max then max=adf end
    end
    print("sumsq = ", sumsq:tonumber())
    print("max = ", max:tonumber())
    for i=1,2 do
      print(i, xi[i], fi[i], yi[i])
    end
  end
  local function w1(i) return 1 end
  local function w0(i) return fi[i] end
  updatec(w1)
  local function wq(i) return 1/polyval(qc,xi[i]) end
  local function wqr(i) return 1/(fi[i]*polyval(qc,xi[i])) end
  for i=1,50 do
    updatec(wqr)
  end
  --local qr = polyroots(qc)
  --m = matrix(ni, d+1)
end

function f(x)
  return 1/mysqrt(x)
end

bc.digits(200)
ratcoeffs(f, 5, 1e-10, 1)

ratfuncs = {}
ratfuncs[-1/2] = {}

ratfuncs[-1/2][5] = {
  coeffs = {
    5.3373368882049743e-02,
    2.1067278541828438e-02,
    6.5015685808105711e-02,
    2.4479827820044367e-01,
    9.4745796107981517e-01,
    4.3977750671373919e+00
  },
  shifts = {
    false,
    2.0859203185118118e-04,
    5.3957612308936130e-03,
    8.1667969249242786e-02,
    1.2077243320062676e+00,
    2.0165690485778253e+01
  }
}

ratfuncs[-1/4] = {}
ratfuncs[-1/4][5] = {
  coeffs = {
    2.4180997518852884e-01,
    2.2934220711468491e-03,
    1.6073585387165145e-02,
    1.1861091867689499e-01,
    8.9673330331604229e-01,
    8.9710997891569946e+00
  },
  shifts = {
    false,
    3.7359430152758880e-04,
    7.7397767594550105e-03,
    1.1449644052137110e-01,
    1.6761943322737960e+00,
    2.8897377449259910e+01
  }
}
